import settings_helper as sh
import dt_helper as dh
import input_helper as ih
from collections import OrderedDict
from contextlib import suppress
from pymongo import MongoClient
from mongo_helper.queries import *


get_setting = sh.settings_getter(__name__)
mongo_url = get_setting('mongo_url')


class Mongo(object):
    def __init__(self, url=None, db='db'):
        """An instance that can execute MongoDB statements

        - url: connection url to a MongoDB
        - db: name of db to use for making queries
        """
        self._client = MongoClient(url)
        self._db = db

    def get_databases(self, system=False):
        """Return a list of database names

        - system: if True, include the system dbs 'admin', 'config', and 'local'
        """
        ignore = []
        if not system:
            ignore = ['admin', 'config', 'local']
        return [
            name
            for name in self._client.list_database_names()
            if name not in ignore
        ]

    def change_database(self, db):
        """Set a different db to use for making queries"""
        self._db = db

    def select_database(self, system=False):
        """Interactively select a db to use for making queries

        - system: if True, include the system dbs 'admin', 'config', and 'local'
          in menu of dbs
        """
        selected = ih.make_selections(
            self.get_databases(system=system),
            prompt='Select database',
            one=True
        )
        if selected:
            self._db = selected

    def get_collections(self, db=None):
        """Return a list of collection names

        - db: if None, use current value in self._db
        """
        db = self._db if db is None else db
        return self._client[db].list_collection_names()

    def get_all_collections_for_all_databases(self, system=False):
        """Return a dict of database names and their collections

        - system: if True, include the system dbs 'admin', 'config', and 'local'
        """
        return {
            db: self.get_collections(db)
            for db in self.get_databases(system=system)
        }

    def db_stats(self):
        """Return a dict of info about the db

        See: https://docs.mongodb.com/manual/reference/command/dbStats/#output
        """
        return self._command('dbStats')

    def coll_stats(self, collection):
        """Return a dict of info about the collection

        See: https://docs.mongodb.com/manual/reference/command/collStats/#output
        """
        return self._command('collStats', collection)

    def server_status(self):
        """Return a dict of info about the server

        See: https://docs.mongodb.com/manual/reference/command/serverStatus/#output
        """
        return self._command('serverStatus')

    def _command(self, *args, **kwargs):
        """Run a db command and return the results"""
        db = self._db
        return self._client[db].command(*args, **kwargs)

    def _insert_one(self, collection, document):
        """Add a document to the collection and return inserted_id

        - document: a dict of info to be inserted
        """
        db = self._db
        result = self._client[db][collection].insert_one(document)
        return result.inserted_id

    def _insert_many(self, collection, list_of_dicts):
        """Add several documents to the collection and return inserted_ids"""
        db = self._db
        result = self._client[db][collection].insert_many(list_of_dicts)
        return result.inserted_ids

    def _find(self, collection, *args, **kwargs):
        """Return a cursor"""
        db = self._db
        return self._client[db][collection].find(*args, **kwargs)

    def _find_one(self, collection, *args, **kwargs):
        """Return an object"""
        db = self._db
        return self._client[db][collection].find_one(*args, **kwargs)

    def _count(self, collection, *args, **kwargs):
        """Return an int"""
        db = self._db
        return self._client[db][collection].count(*args, **kwargs)

    def _aggregate(self, collection, *args, **kwargs):
        """Return a cursor"""
        db = self._db
        return self._client[db][collection].aggregate(*args, **kwargs)

    def _create_index(self, collection, keys, unique=False, ttl=None,
                      sparse=False, background=False, **kwargs):
        """Create an index on the collection

        - keys: list of 2-item tuples where first item is a field name
          and second item is a direction (1 for ascending, -1 for descending)
            - keys can also be a single key (string) or list of strings
        - unique: if True, create a uniqueness constraint
        - ttl: int representing "time to live" (in seconds) for documents in
          the collection
        - sparse: if True, only index documents that contain the indexed field
        - background: if True, create the index in the background
        """
        kwargs['unique'] = unique
        kwargs['sparse'] = sparse
        kwargs['background'] = background
        if ttl is not None:
            kwargs['expireAfterSeconds'] = ttl
        db = self._db
        return self._client[db][collection].create_index(keys, **kwargs)

    def _drop_index(self, collection, name):
        """Drop an index from the collection"""
        db = self._db
        return self._client[db][collection].drop_index(name, **kwargs)

    def _index_information(self, collection):
        """Return a dict of info about indexes on collection"""
        db = self._db
        try:
            index_info = self._client[db][collection].index_information()
        except:
            index_info = {}
        return index_info

    def last_obj(self, collection, timestamp_field='_id', **kwargs):
        """Return last object inserted to collection

        - timestamp_field: name of timestamp field to sort on
        - kwargs: passed to `self._find_one`
        """
        if 'sort' not in kwargs:
            kwargs['sort'] = [(timestamp_field, -1)]

        return self._find_one(collection, **kwargs)

    def first_obj(self, collection, timestamp_field='_id', **kwargs):
        """Return first object inserted to collection

        - timestamp_field: name of timestamp field to sort on
        - kwargs: passed to `self._find_one`
        """
        if 'sort' not in kwargs:
            kwargs['sort'] = [(timestamp_field, 1)]

        return self._find_one(collection, **kwargs)

    def obj_id_set(self, collection, match):
        """Return set of ObjectIds for match

        - match: dictionary representing the documents to match
        """
        return set([
            x['_id']
            for x in self._find(collection, match, projection=['_id'])
        ])

    def _build_pipeline(self, match=None, group_by=None, timestamp_field='_id',
                    unwind=None, include_array_index=False, projection=None,
                    limit=None, to_set=None, to_list=None, to_sum=None,
                    group_action=None, include_condition=None, verbose=False):
        """Return a list with pipeline stages, to be passed to aggregate method

        - match: dictionary representing the "match stage"
        - group_by: list of keys to group by
        - timestamp_field: name of timestamp field to sort on (if 'limit' != None)
        - unwind: list of keys unwind (key value in document should be an array)
        - include_array_index: if True and 'unwind' is specified, the index of each
          unwound item will be included
        - projection: list of keys to project
        - limit: max number of items
        - to_set: list of keys, where each key will have its values added to a
          set for each unique group
        - to_list: list of keys, where each key will have its values added to a
          list for each unique group
        - to_sum: list of keys, where each key will have its values summed
          for each unique group
        """
        pipeline = []

        if match:
            pipeline.append({'$match': match})

        if limit:
            pipeline.append({'$sort': {timestamp_field: -1}})
            pipeline.append({'$limit': limit})

        if unwind:
            if include_array_index is True:
                for key in unwind:
                    pipeline.append({'$unwind': {
                        'path': '${}'.format(key),
                        'includeArrayIndex': '{}_idx'.format(key)
                    }})
            else:
                for key in unwind:
                    pipeline.append({'$unwind': '${}'.format(key)})

        if projection:
            pipeline.append({'$project': {k: '${}'.format(k) for k in projection}})

        if group_by:
            group = {
                '$group': {
                    '_id': {k: '${}'.format(k) for k in group_by},
                    'count': {'$sum': 1},
                }
            }
        else:
            group = {'$group': {}}

        if to_set:
            group['$group'].update({
                k: {'$addToSet': '${}'.format(k)} for k in to_set
            })

        if to_list:
            group['$group'].update({
                k: {'$push': '${}'.format(k)} for k in to_list
            })

        if to_sum:
            group['$group'].update({
                k: {'$sum': '${}'.format(k)} for k in to_sum
            })

        if group != {'$group': {}}:
            pipeline.append(group)
            pipeline.append({'$sort': {'count': -1}})

        return pipeline

    def _explain_pipeline(self, collection, pipeline):
        """Return a dict"""
        return self._command('aggregate', collection, pipeline=pipeline, explain=True)

    def ez_pipeline(self, collection, match, group_by, timestamp_field='_id',
                    unwind=None, include_array_index=False, projection=None,
                    limit=None, to_set=None, to_list=None, to_sum=None,
                    group_action=None, include_condition=None, verbose=False):
        """Build/run an aggregation pipeline to group and count data

        - collection: name of collection
        - match: dictionary representing the "match stage"
        - group_by: list of keys to group by
        - timestamp_field: name of timestamp field to sort on (if 'limit' != None)
        - unwind: list of keys unwind (key value in document should be an array)
        - include_array_index: if True and 'unwind' is specified, the index of each
          unwound item will be included
        - projection: list of keys to project
        - limit: max number of items
        - to_set: list of keys, where each key will have its values added to a
          set for each unique group
        - to_list: list of keys, where each key will have its values added to a
          list for each unique group
        - to_sum: list of keys, where each key will have its values summed
          for each unique group

        (After aggregation)
        - group_action: callable that will be mapped over each grouped item
        - include_condition: callable returning a bool to determine if a grouped
          item will be included in returned data
        - verbose: if True, print the generated pipeline command

        Return a dictionary with keys 'counts', 'data', 'total', 'group_by',
        'duration', 'pipeline', and 'total_percent'.
        """
        _start = dh.utc_now_localized()
        pipeline = self._build_pipeline(
            match=match, group_by=group_by, timestamp_field=timestamp_field,
            unwind=unwind, include_array_index=include_array_index,
            projection=projection, limit=limit, to_set=to_set, to_list=to_list,
            to_sum=to_sum
        )

        if verbose:
            from pprint import pprint
            pprint(pipeline)

        cursor = self._aggregate(collection, pipeline)
        counts = []
        data = {}
        total = 0

        if include_condition is None:
            include_condition = lambda x: True
        elif not callable(include_condition):
            include_condition = lambda x: True

        for item in cursor:
            count = item.pop('count')
            total += count

            if include_condition(item):
                the_group = tuple([item['_id'].get(k, '') for k in group_by])
                with suppress(TypeError):
                    item = group_action(item)
                counts.append((the_group, count))
                data[the_group] = item

        final_counts = {'data': []}
        for name, count in counts:
            final_counts['data'].append((name, count, count/total))

        final_data = OrderedDict()
        for x in final_counts['data']:
            key = x[0]
            final_data[key] = data[key]

        total_percent = sum([x[-1] for x in final_counts['data']])

        _end = dh.utc_now_localized()

        results = {
            'counts': final_counts,
            'data': final_data,
            'total': total,
            'total_percent': total_percent,
            'group_by': tuple(group_by),
            'duration': (_end - _start).total_seconds(),
            'pipeline': pipeline,
        }
        return results
